
var fs = require('fs'),
    path = require('path'),
    gruntMod = require('grunt'),
    helpers = require('./helpers');


var compiler = {};


/**
 * Perform validations for given options
 *
 * @param  {Object} options [description]
 * @return {boolean} boolean.
 */
compiler.validateOpts = function validate( options ) {
  // check for closure compiler file
  var compilerJar = options.compilerFile;
  //
  // check compilerJar's existence
  //
  if (!gruntMod.file.isFile( compilerJar )) {
    helpers.log.error('compiler Jar file path not valid: ' + compilerJar.red);
    return false;
  }
  return true;
};

/**
 * Validates a grunt file object
 * @param  {Object} fileObj Grunt file object
 * @return {boolean} yez or noz.
 */
compiler.validateFile = function validateFile( fileObj ) {

  var warningMsg = 'WARNING :: No "src" files were defined';
  //
  // Check for js source files
  //
  if ( !fileObj.src || 0 === gruntMod.file.expand( fileObj.src )) {
    helpers.log.warn(warningMsg);
  }

  return true;
};


/**
 * Prepare and compile the compiler command we will execute
 *
 * @param {Object} options The options.
 * @param {Object} fileObj The file object to be compiled.
 * @return {string|boolean} boolean false if we failed,
 *   command string if all ok.
 */
compiler.compileCommand = function compileCommand( options, fileObj ) {

  var cmd = 'java ' +
    (options.javaFlags ? options.javaFlags.join(' ') : '') +
    // optional speed optimizations
    // d32 as found by Igor Minar http://than.pol.as/NfzB
    (options.d32 ? ' -client -d32 ' : '') +
    (options.TieredCompilation ? ' -server -XX:+TieredCompilation ' : '') +
    ' -jar "' + options.compilerFile + '" ';

  //
  // check for js files
  //
  var src = gruntMod.file.expand({nonull: true}, fileObj.src || []);
  if (src.length > 0) {
    cmd += helpers.makeParam(src, '--js');
  }

  //
  // Process output file
  //
  var dest = fileObj.dest;
  if('string' === typeof(dest)) {
    dest = gruntMod.template.process(dest);
  }

  // Check if there are any files to write out. If not, make up an empty command.
  if (src.length === 0) {
    helpers.log.warn('Either no src files selected or all filtered out. Writing empty file to ' + dest.cyan);
    // Command sub object for internal representation of commands that aren't executed in the shell
    cmd = {
      type: 'empty',
      dest: dest
    };
    return cmd;
  }
  // check if output file is defined
  if (dest && dest.length) {

    // create the directories if not there already
    gruntMod.file.mkdir(path.dirname( dest ));

    // if we want to check the source files timestamps enter this loop.
    if (options.checkModified && fs.existsSync( dest )) {
      var doCompile = false;
      var outMtime = fs.lstatSync( dest ).mtime;

      for(var i = 0, len = src.length ; i < len ; i++) {
        if ( fs.lstatSync( src[i] ).mtime > outMtime ) {
          doCompile = true;
          break;
        }
      }
      if (!doCompile) {
        helpers.log.info('Skipping ' + src.blue + ' (Not modified)');
        return false;
      }
    }

    cmd += ' --js_output_file=' + dest;
  }

  //
  // start digging on compiler options
  //

  // FIXME FIXME remove when gruntjs/grunt#738 is resolved
  // get compiler options
  var opts = gruntMod.util._.extend({},
    options.compilerOpts,
    fileObj.TEMPcompilerOpts);

  for(var directive in opts) {

    // look for 'externs' special case
    if ('externs' === directive) {
      opts[directive] = gruntMod.file.expand( opts[directive] );
    }

    cmd += helpers.makeParam( opts[directive], '--' + directive );
  }


  return cmd;

};

module.exports = compiler;
